/* 
 * command-io.c : Command Persistence (aka Macros, a form of Change Tracking)
 * 
 * For each command type in commands.c (commands.h provides the basic list),
 * a cmdio_write function is written here to write relevant command details
 * to an xml log file.
 * In the (not too distant) future, read functions will be added so that a
 * log file can be read and commands executed...similar to macros?
 *
 * Menu items in "Tools->Change Tracking" are used to start and stop recording
 *
 * TODO:
 *	- TEST newly added functions
 *	- Pango markup
 *	- FIX cmdio_write_paste_cut (GnmParsePos causing problems)
 *	- MergeCells isn't storing the selection correctly
 *	- Disable Menu Options based on wb->change_tracker.enabled
 *	- Give Start... menu item a dialog (choose where to save log)
 */

#include <gnumeric-config.h>
#include <glib/gi18n-lib.h>
#include "gnumeric.h"
#include "command-io.h"

#include "application.h"
#include "sheet.h"
#include "sheet-view.h"
#include "sheet-style.h"
#include "gnm-format.h"
#include "format-template.h"
#include "command-context.h"
#include "wbc-gtk.h"
#include "wbc-gtk-impl.h"
#include "libgnumeric.h"
#include "workbook-control.h"
#include "workbook-view.h"
#include "workbook-priv.h"
#include "ranges.h"
#include "sort.h"
#include "dependent.h"
#include "value.h"
#include "expr.h"
#include "expr-name.h"
#include "cell.h"
#include "sheet-merge.h"
#include "parse-util.h"
#include "print-info.h"
#include "clipboard.h"
#include "selection.h"
#include "colrow.h"
#include "style-border.h"
#include "auto-correct.h"
#include "sheet-autofill.h"
#include "mstyle.h"
#include "search.h"
#include "gutils.h"
#include "gui-util.h"
#include "sheet-object-cell-comment.h"
#include "sheet-object-widget.h"
#include "sheet-object.h"
#include "sheet-object-impl.h"
#include "sheet-object-graph.h"
#include "sheet-control.h"
#include "style-color.h"
#include "auto-format.h"
#include "tools/dao.h"
#include "gnumeric-gconf.h"
#include "scenarios.h"
#include "data-shuffling.h"
#include "tools/tabulate.h"
#include "xml-sax.h"

#include <gsf/gsf-impl-utils.h>
#include <gsf/gsf-doc-meta-data.h>
#include <gsf/gsf-output-stdio.h>
#include <gsf/gsf-output-memory.h>
#include <string.h>
#include <goffice/graph/gog-graph.h>
#include <goffice/utils/go-glib-extras.h>
#include <goffice/utils/go-pango-extras.h>
#include <goffice/utils/go-file.h>
#include <pango/pango-attributes.h>

void
cmdio_start (WBCGtk *wbcg)
{
	char *filename;
	GsfOutput *file;
	
	WorkbookControl *wbc = WORKBOOK_CONTROL (wbcg);
	Workbook *wb = wb_control_get_workbook(wbc);

	g_print("ENTER cmdio_start\n");

	if (!wb->change_tracker.enabled) {
		filename = g_strdup_printf ("%s-commands.log",
				g_path_get_basename (go_doc_get_uri (GO_DOC (wb))));
		g_print("Log Filename: [%s]\n", filename);
		file = gsf_output_stdio_new (filename, NULL);
		if (file) {
			GsfXMLOut *output = gsf_xml_out_new (file);
			
			wb->change_tracker.enabled = TRUE;
			wb->change_tracker.file = file;
			wb->change_tracker.output = output;
		}
		g_free (filename);
	}
}

void
cmdio_stop (WBCGtk *wbcg)
{
	GsfOutput *file;
	WorkbookControl *wbc = WORKBOOK_CONTROL (wbcg);
	Workbook *wb = wb_control_get_workbook(wbc);

	g_print("ENTER cmdio_stop\n");
	
	if (wb->change_tracker.enabled == TRUE) {
		file = wb->change_tracker.file;
		if (file) {
			g_object_unref (G_OBJECT (wb->change_tracker.output));
			gsf_output_close (GSF_OUTPUT (file));
			g_object_unref (file);
		}
		wb->change_tracker.enabled = FALSE;
	}
	/* out.seek (end);  action.write (out); end = out.tell();   out.endElement ("Log"); */
}

void
cmdio_write_sheet (GsfXMLOut *output, char const *id, Sheet *sheet)
{
	g_print("\tENTER cmdio_write_sheet\n");
	g_return_if_fail (IS_SHEET (sheet));
	
	gsf_xml_out_start_element (output, "Sheet");
	if (id != NULL) {
		gsf_xml_out_add_cstr (output, "ID", id);
	}
	gsf_xml_out_add_cstr (output, NULL, sheet->name_unquoted);
	gsf_xml_out_end_element (output);
}

void
cmdio_write_position (GsfXMLOut *output, char const *id, int col, int row)
{
	g_print("\tENTER cmdio_write_position\n");
	gsf_xml_out_start_element (output, "Position");
	if (id != NULL) {
		gsf_xml_out_add_cstr (output, "ID", id);
	}	
	gsf_xml_out_add_int (output, "Col", col);
	gsf_xml_out_add_int (output, "Row", row);
	gsf_xml_out_end_element (output); /* </Position> */
}

void
cmdio_write_range (GsfXMLOut *output, char const *id, GnmRange const *r)
{
	g_print("\tENTER cmdio_write_range\n");
	gsf_xml_out_start_element (output, "Range");
	if (id != NULL) {
		gsf_xml_out_add_cstr (output, "ID", id);
	}
	cmdio_write_position (output, "Start", r->start.col, r->start.row);
	cmdio_write_position (output, "End", r->end.col, r->end.row);
	gsf_xml_out_end_element (output);
}

void
cmdio_write_markup (GsfXMLOut *output, PangoAttrList const *markup)
{
	/* GSList *ptr; */

	g_print("\tENTER cmdio_write_markup\n");

	gsf_xml_out_start_element (output, "PangoMarkup");
	
#if 0	
	TODO: make all of this work...
	gsf_xml_out_add_uint (output, "RefCount", markup->ref_count);

	for (ptr = markup->attributes; ptr != NULL; ptr = ptr->next) {
		PangoAttribute *attr = ptr->data;
		
		gsf_xml_out_start_element (output, "PangoAttribute");
		gsf_xml_out_add_cstr_unchecked (output, "Type",
				pango_attr_type_get_name(attr->klass->type));
		gsf_xml_out_add_uint (output, "Start", attr->start_index);
		gsf_xml_out_add_uint (output, "End", attr->end_index);
		gsf_xml_out_add_cstr (output, NULL, attr->klass->value);
		gsf_xml_out_end_element (output); /* </PangoAttribute> */
	}
#endif

	gsf_xml_out_end_element (output); /* </PangoMarkup> */
}

void
cmdio_write_sheetview (GsfXMLOut *output, SheetView const *sv)
{
	/* TODO: need to store anything else?? */
	GSList *ptr;

	g_print("\tENTER cmdio_write_sheetview\n");

	gsf_xml_out_start_element (output, "SheetView");
	gsf_xml_out_add_int (output, "FirstTabCol", sv->first_tab_col);
	
	cmdio_write_sheet (output, NULL, sv->sheet);
	
	gsf_xml_out_start_element (output, "Selections");
	for (ptr = sv->selections; ptr != NULL ; ptr = ptr->next) {
		GnmRange const *r = ptr->data;
		cmdio_write_range (output, NULL, r);
	}
	gsf_xml_out_end_element (output); /* </Selections> */
	
	cmdio_write_position (output, "editPos", sv->edit_pos.col, sv->edit_pos.row);
	cmdio_write_position (output, "editPosReal", sv->edit_pos_real.col, sv->edit_pos_real.row);

	gsf_xml_out_end_element (output); /* </SheetView> */
}

void
cmdio_write_parse_pos (GsfXMLOut *output, GnmParsePos const *pos)
{
	/* FIXME: this is causing a segfault when a selection is dragged */
	g_print("\tENTER cmdio_write_parse_pos\n");

	g_return_if_fail (IS_SHEET (pos->sheet));
	
	gsf_xml_out_start_element (output, "ParsePos");
	cmdio_write_sheet (output, NULL, pos->sheet);
	cmdio_write_position (output, NULL, pos->eval.col, pos->eval.row);
	/* TODO pos->wb */
	gsf_xml_out_end_element (output);
}

void
cmdio_write_paste_target	(GsfXMLOut *output, GnmPasteTarget const *pt)
{
	g_print("\tENTER cmdio_write_paste_target\n");
	gsf_xml_out_start_element (output, "PasteTarget");
	gsf_xml_out_add_int (output, "PasteFlags", pt->paste_flags);
	cmdio_write_sheet (output, NULL, pt->sheet);
	cmdio_write_range (output, NULL, &pt->range);
	gsf_xml_out_end_element (output); /* </PasteTarget> */
}

void
cmdio_write_style		(GsfXMLOut *output, WorkbookControl *wbc, GnmStyle const *style)
{
	GnmOutputXML state;
	
	state.wb_view	= wb_control_view (wbc);
	state.wb	= wb_control_get_workbook (wbc);
	state.sheet	= NULL;
	state.output	= output;
	state.convs	= gnm_xml_io_conventions ();
	state.expr_map  = g_hash_table_new (g_direct_hash, g_direct_equal);
	state.cell_str  = g_string_new (NULL);
	
	xml_write_style (&state, style); /* see xml-sax-write.c:378 */
}

void
cmdio_write_template_member (GsfXMLOut *output, WorkbookControl *wbc, TemplateMember const *tm)
{
	gsf_xml_out_start_element (output, "TemplateMember");
	gsf_xml_out_add_int (output, "Repeat", tm->repeat);
	gsf_xml_out_add_int (output, "Skip", tm->skip);
	gsf_xml_out_add_int (output, "Edge", tm->edge);
	/* gsf_xml_out_add_enum (output, "FreqDirection", FreqDirection, tm->direction); */
	gsf_xml_out_add_int (output, "FreqDirection", tm->direction);	
	
	gsf_xml_out_start_element (output, "RowInfo");
	gsf_xml_out_add_int (output, "Offset", tm->row.offset);
	gsf_xml_out_add_int (output, "OffsetGravity", tm->row.offset_gravity);
	gsf_xml_out_add_int (output, "Size", tm->row.size);
	gsf_xml_out_end_element (output); /* </RowInfo> */

	gsf_xml_out_start_element (output, "ColInfo");
	gsf_xml_out_add_int (output, "Offset", tm->col.offset);
	gsf_xml_out_add_int (output, "OffsetGravity", tm->col.offset_gravity);
	gsf_xml_out_add_int (output, "Size", tm->col.size);
	gsf_xml_out_end_element (output); /* </ColInfo> */

	cmdio_write_style (output, wbc, tm->mstyle);

	gsf_xml_out_end_element (output); /* </TemplateMember> */
}

void
cmdio_write_sheet_object_view	(GsfXMLOut *output, SheetObjectView const *sov)
{
	gsf_xml_out_start_element (output, "SheetObjectView");
	/* TODO: need to figure this out */
	gsf_xml_out_end_element (output);
}

void
cmdio_write_sheet_object	(GsfXMLOut *output, SheetObject const *so)
{
	GList *l;
	gsf_xml_out_start_element (output, "SheetObject");
	gsf_xml_out_add_int (output, "ObjectFlags", so->flags);

	cmdio_write_sheet (output, NULL, so->sheet);
	
	for (l = so->realized_list; l != NULL; l = l->next) {
		SheetObjectView *view = SHEET_OBJECT_VIEW (l->data);
		cmdio_write_sheet_object_view (output, view);
	}
	
	cmdio_write_anchor (output, &so->anchor);
		
	gsf_xml_out_end_element (output); /* </SheetObject> */
}

void
cmdio_write_anchor	(GsfXMLOut *output, SheetObjectAnchor const *anchor)
{
	/* TODO: need a better way to do the for loops here */
	int i;

	gsf_xml_out_start_element (output, "SheetObjectAnchor");

	gsf_xml_out_start_element (output, "DrawingAnchor");
	gsf_xml_out_add_int (output, "Dir", anchor->base.direction);

	for (i = 0; i < 4; i++) {
		gsf_xml_out_simple_int_element (output, "PosPts", anchor->base.pos_pts[i]);
	}
	gsf_xml_out_end_element (output); /* </DrawingAnchor> */
	
	cmdio_write_range (output, NULL, &anchor->cell_bound);
	
	for (i = 0; i < 4; i++) {
		gsf_xml_out_simple_float_element (output, "Offset", anchor->offset[i], -1);
	}
	
	gsf_xml_out_end_element (output); /* </SheetObjectAnchor> */
}

/******************************************************************************/

void
cmdio_write_set_text	(WorkbookControl *wbc, Sheet *sheet,
	                 GnmCellPos const *pos, char const *new_text,
        	         PangoAttrList *markup) 
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER cmdio_write_set_text\n");	

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;

		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdSetText");
	
		cmdio_write_sheet (output, NULL, sheet);
		cmdio_write_position (output, NULL, pos->col, pos->row);
		gsf_xml_out_simple_element (output, "NewText", new_text);
		cmdio_write_markup (output, markup);
	
		gsf_xml_out_end_element (output); /* </Command> */
	}
}


void
cmdio_write_area_set_text	(WorkbookControl *wbc, SheetView *sv,
				 char const *text, gboolean as_array)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_area_set_text\n");
	
	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;

		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdAreaSetText");
		gsf_xml_out_add_bool (output, "AsArray", as_array);
		
		cmdio_write_sheetview (output, sv);
		gsf_xml_out_simple_element (output, "Text", text);
	
		gsf_xml_out_end_element (output); /* </Command> */
	}
}


void
cmdio_write_create_data_table	(WorkbookControl *wbc,
				 Sheet *sheet, GnmRange const *r,
				 char const *col_input, char const *row_input)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_create_data_table\n");
	
	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;

		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdCreateDataTable");
		gsf_xml_out_add_cstr_unchecked (output, "ColInput", col_input);
		gsf_xml_out_add_cstr_unchecked (output, "RowInput", row_input);

		cmdio_write_sheet (output, NULL, sheet);
		cmdio_write_range (output, NULL, r);
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_ins_del_colrow	(WorkbookControl *wbc, Sheet *sheet,
				 gboolean is_cols, gboolean is_insert,
				 int index, int count)
{
	const char *insdel, *colrow, *cmd_type;
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_ins_del_colrow\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		insdel = is_insert ? "Ins" : "Del";
		colrow = is_cols ? "Col" : "Row";
		cmd_type = g_strdup_printf ("Cmd%s%s", insdel, colrow);
			
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr (output, "Type", cmd_type);
		gsf_xml_out_add_int (output, "Index", index);
		gsf_xml_out_add_int (output, "Count", count);
		
		cmdio_write_sheet (output, NULL, sheet);
			
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_resize_colrow	(WorkbookControl *wbc, Sheet *sheet,
				 gboolean is_cols, ColRowIndexList *selection,
				 int new_size)
{
	ColRowIndexList *ptr;
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_resize_colrow\n");
	
	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;

		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdResizeColRow");
		gsf_xml_out_add_bool (output, "IsCols", is_cols);
		gsf_xml_out_add_int (output, "NewSize", new_size);
		
		cmdio_write_sheet (output, NULL, sheet);

		gsf_xml_out_start_element (output, "ColRowIndexList");
		for (ptr = selection; ptr != NULL ; ptr = ptr->next) {
			ColRowIndex const *index = ptr->data;
			gsf_xml_out_start_element (output, "ColRowIndex");
			gsf_xml_out_add_int (output, "First", index->first);
			gsf_xml_out_add_int (output, "Last", index->last);
			gsf_xml_out_end_element (output);
		}
		gsf_xml_out_end_element (output); /* </ColRowIndexList> */
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_paste_cut		(WorkbookControl *wbc,
				 GnmExprRelocateInfo const *info,
				 gboolean move_selection, char *cmd_descriptor)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_paste_cut\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdPasteCut");
		gsf_xml_out_add_bool (output, "MoveSelection", move_selection);
		gsf_xml_out_add_cstr_unchecked (output, "Descriptor", cmd_descriptor);
		
		gsf_xml_out_start_element (output, "ExprRelocateInfo");
		gsf_xml_out_add_int (output, "ColOffset", info->col_offset);
		gsf_xml_out_add_int (output, "RowOffset", info->row_offset);
		/* TODO: enum info->reloc_type
		 * 	 does it work to do it as an int? (see below)
		 *
		 *	if not:
		 *	gsf_xml_out_add_enum (GsfXMLOut *xout, char const *id, GType etype, gint val)
		 *	 but what GType to use?
		 */
		gsf_xml_out_add_int (output, "RelocType", info->reloc_type);
		
		/* FIXME: 
		 * This is causing a segfault -- can ParsePos be NULL??
		 *
 		 * cmdio_write_parse_pos (output, &info->pos);
 		 * g_print("\tParsePos done\n");
 		 */
		
		cmdio_write_range (output, "Origin", &info->origin);
		cmdio_write_sheet (output, "Origin", info->origin_sheet);
		cmdio_write_sheet (output, "Target", info->target_sheet);
		gsf_xml_out_end_element (output); /* </ExprRelocateInfo> */
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_paste_copy	(WorkbookControl *wbc,
			 GnmPasteTarget const *pt, GnmCellRegion const *cr)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_paste_copy\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdPasteCopy");

		cmdio_write_paste_target (output, pt);
		
		/* cmdio_write_cell_region (output, cr);
		 * TODO: make this work
		 * See xml-sax-write.c:1343 for gnm_cellregion_to_xml
		 *
		GsfOutputMemory *outmem  = gnm_cellregion_to_xml (cr);
		if (outmem) {
			* write to file...hopefully 
			* see gui-clipboard.c:699 & sheet-control-gui.c:3588
			* gsf_output_memory_write (file, outmem->capacity,
			*			 gsf_output_memory_get_bytes (outmem));
			* 
			* gsf_output_memory_get_bytes (output);
			* gsf_output_size (GSF_OUTPUT (output));	size_t ??
			*
			g_object_unref (outmem);
		}
		 */
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_sort	(WorkbookControl *wbc, GnmSortData const *data)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_sort\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdSort");
		
		gsf_xml_out_start_element (output, "SortData");
		gsf_xml_out_add_int (output, "NumClause", data->num_clause);
		gsf_xml_out_add_bool (output, "Top", data->top);
		gsf_xml_out_add_bool (output, "RetainFormats", data->retain_formats);
						
		cmdio_write_sheet (output, NULL, data->sheet);
		cmdio_write_range (output, NULL, data->range);
		
		gsf_xml_out_start_element (output, "SortClause");
		gsf_xml_out_add_int (output, "Offset", data->clauses->offset);
		gsf_xml_out_add_bool (output, "Asc", data->clauses->asc);
		gsf_xml_out_add_bool (output, "Cs", data->clauses->cs);
		gsf_xml_out_add_bool (output, "Val", data->clauses->val);
		gsf_xml_out_end_element (output); /* </SortClause> */
		
		gsf_xml_out_simple_element (output, "Locale", data->locale);
		
		gsf_xml_out_end_element (output); /* </SortData> */
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void 
cmdio_write_autofill	(WorkbookControl *wbc, Sheet *sheet,
			 gboolean default_increment,
			 int base_col, int base_row,
			 int w, int h, int end_col, int end_row,
			 gboolean inverse_autofill)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_autofill\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdAutofill");
		gsf_xml_out_add_bool (output, "DefaultIncrement", default_increment);
		gsf_xml_out_add_int (output, "BaseCol", base_col);
		gsf_xml_out_add_int (output, "BaseRow", base_row);
		gsf_xml_out_add_int (output, "Width", w);
		gsf_xml_out_add_int (output, "Height", h);
		gsf_xml_out_add_int (output, "EndCol", end_col);
		gsf_xml_out_add_int (output, "EndRow", end_row);
		gsf_xml_out_add_bool (output, "InverseAutofill", inverse_autofill);

		cmdio_write_sheet (output, NULL, sheet);
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_copyrel	(WorkbookControl *wbc,
			 int dx, int dy,
			 char const *name)
{
	/* TODO: test this function...how is it invoked? */
	Workbook *wb = wb_control_get_workbook(wbc);

	g_print("ENTER write_copyrel\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdCopyRel");
		gsf_xml_out_add_int (output, "Dx", dx);
		gsf_xml_out_add_int (output, "Dy", dy);

		gsf_xml_out_simple_element (output, "Name", name);

		gsf_xml_out_end_element (output); /* </Command> */
	}
}

/***************************************************************************
 * The following cmdio_write_selection_ functions need to get the selection 
 * (...as well as the current sheet?) from the calling control
 */
void
cmdio_write_selection_format	(WorkbookControl *wbc,
				 GnmStyle *style, GnmBorder **borders,
				 char const *opt_translated_name)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_selection_format\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
				
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdFormat");
		
		/* TODO: do we need to write_sheet separately? */
		cmdio_write_sheetview (output, wb_control_cur_sheet_view(wbc));
	
		if (style != NULL)
			cmdio_write_style (output, wbc, style);
	
		if (borders != NULL) {
		/* TODO:
		 * see commands.c:1930 -- need a for loop
		 *	gsf_xml_out_start_element (output, "Borders");
		 *	gsf_xml_out_simple_int_element (output, "StyleBorderType",
		 *					borders->line_type);
		 *	* gsf_xml_out_add_enum (output, "StyleBorderType",
		 *				GnmStyleBorderType, borders->line_type); *
		 *	gsf_xml_out_start_element (output, "Color");
		 *	gsf_xml_out_add_color (output, NULL, borders->color);
		 *	gsf_xml_out_end_element (output); * </StyleBorderType> *
		 *	gsf_xml_out_simple_int_element (output, "BeginMargin", borders->begin_margin);
		 *	gsf_xml_out_simple_int_element (output, "EndMargin", borders->end_margin);
		 *	gsf_xml_out_simple_int_element (output, "Width", borders->width);
		 *	* TODO: need 'private' items?? -- style-border.h *
		 *	gsf_xml_out_end_element (output); * </Borders> *
		 */
		}
		
		if (opt_translated_name != NULL)
			gsf_xml_out_simple_element (output, "TranslatedName",
							opt_translated_name);

		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_selection_clear	(WorkbookControl *wbc, int clear_flags)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_selection_clear\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdClear");
		gsf_xml_out_add_int (output, "ClearFlags", clear_flags);
		
		/* TODO: do we need to write_sheet separately? */
		cmdio_write_sheetview (output, wb_control_cur_sheet_view(wbc));

		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_selection_colrow_hide	(WorkbookControl *wbc,
					 gboolean is_cols, gboolean visible)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_selection_colrow_hide\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdColRowHide");
		gsf_xml_out_add_bool (output, "IsCols", is_cols);
		gsf_xml_out_add_bool (output, "Visible", visible);
		
		/* TODO: do we need to write_sheet separately? */
		cmdio_write_sheetview (output, wb_control_cur_sheet_view(wbc));

		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_selection_outline_change	(WorkbookControl *wbc,
					 gboolean is_cols, int index, int depth)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_selection_outline_change\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdOutlineChange");
		gsf_xml_out_add_bool (output, "IsCols", is_cols);
		gsf_xml_out_add_int (output, "Index", index);
		gsf_xml_out_add_int (output, "Depth", depth);
		
		/* TODO: do we need to write_sheet separately? */
		cmdio_write_sheetview (output, wb_control_cur_sheet_view(wbc));

		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_selection_group	(WorkbookControl *wbc,
				 gboolean is_cols, gboolean group)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_selection_group\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdGroup");
		gsf_xml_out_add_bool (output, "IsCols", is_cols);
		gsf_xml_out_add_bool (output, "Group", group);
				
		/* TODO: do we need to write_sheet separately? */
		cmdio_write_sheetview (output, wb_control_cur_sheet_view(wbc));

		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_selection_autoformat	(WorkbookControl *wbc, GnmFormatTemplate *ft)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_selection_autoformat\n");

	if (wb->change_tracker.enabled == TRUE) {
		GSList *ptr;
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdAutoFormat");
		
		/* TODO: do we need to write_sheet separately? */
		cmdio_write_sheetview (output, wb_control_cur_sheet_view(wbc));

		gsf_xml_out_start_element (output, "FormatTemplate");
		gsf_xml_out_add_bool (output, "Number", ft->number);
		gsf_xml_out_add_bool (output, "Border", ft->border);
		gsf_xml_out_add_bool (output, "Font", ft->font);
		gsf_xml_out_add_bool (output, "Patterns", ft->patterns);
		gsf_xml_out_add_bool (output, "Alignment", ft->alignment);
		
		gsf_xml_out_simple_element (output, "Filename", ft->filename);
		gsf_xml_out_simple_element (output, "Author", ft->author);
		gsf_xml_out_simple_element (output, "Name", ft->name);
		gsf_xml_out_simple_element (output, "Description", ft->description);
		
		gsf_xml_out_start_element (output, "Edges");
		gsf_xml_out_add_bool (output, "Left", ft->edges.left);
		gsf_xml_out_add_bool (output, "Right", ft->edges.right);
		gsf_xml_out_add_bool (output, "Top", ft->edges.top);
		gsf_xml_out_add_bool (output, "Bottom", ft->edges.bottom);
		gsf_xml_out_end_element (output); /* </Edges> */
		
		/* TODO: fix the int casting below (i think this will break)... */
		gsf_xml_out_simple_int_element (output, "FormatTemplateCategory",
						(int)ft->category);
		/* gsf_xml_out_add_enum (output, "FormatTemplateCategory",
					FormatTemplateCategory, ft->category); */
	
		gsf_xml_out_start_element (output, "TemplateMembers");
		for (ptr = ft->members; ptr != NULL ; ptr = ptr->next) {
			TemplateMember const *tm = ptr->data;
			cmdio_write_template_member (output, wbc, tm);
		}
		gsf_xml_out_end_element (output); /* </TemplateMembers> */
		
		/* TODO: add FormatTemplate private members?? */
		
		gsf_xml_out_end_element (output); /* </FormatTemplate> */	
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

/******************************************/

void
cmdio_write_global_outline_change	(WorkbookControl *wbc, gboolean is_cols, int depth)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_global_outline_change\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdGlobalOutlineChange");
		gsf_xml_out_add_bool (output, "IsCols", is_cols);
		gsf_xml_out_add_int (output, "Depth", depth);
		
		cmdio_write_sheetview (output, wb_control_cur_sheet_view (wbc));
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_merge_cells		(WorkbookControl *wbc, Sheet *sheet,
				 GSList const *selection, gboolean center)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_merge_cells\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdMergeCells");
		gsf_xml_out_add_bool (output, "Center", center);
		
		cmdio_write_sheet (output, NULL, sheet);
		
		gsf_xml_out_start_element (output, "Selection");
		for ( ; selection != NULL ; selection = selection->next) {
			cmdio_write_range (output, NULL, selection->data);
		}
		gsf_xml_out_end_element (output); /* </Selection> */
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_unmerge_cells	(WorkbookControl *wbc, Sheet *sheet,
				 GSList const *selection)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_unmerge_cells\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdUnmergeCells");
		
		cmdio_write_sheet (output, NULL, sheet);
		
		gsf_xml_out_start_element (output, "Selection");
		for ( ; selection != NULL ; selection = selection->next) {
			cmdio_write_range (output, NULL, selection->data);
		}
		gsf_xml_out_end_element (output); /* </Selection> */
				
		gsf_xml_out_end_element (output); /* </Command> */
	}

}

void
cmdio_write_search_replace	(WorkbookControl *wbc, GnmSearchReplace *sr)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_search_replace\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdSearchReplace");
		
		gsf_xml_out_start_element (output, "SearchReplace");
		gsf_xml_out_add_bool (output, "Query", sr->query);
		gsf_xml_out_add_bool (output, "SearchStrings", sr->search_strings);
		gsf_xml_out_add_bool (output, "SearchOtherValues", sr->search_other_values);
		gsf_xml_out_add_bool (output, "SearchExpressions", sr->search_expressions);
		gsf_xml_out_add_bool (output, "SearchExpressionResults", sr->search_expression_results);
		gsf_xml_out_add_bool (output, "SearchComments", sr->search_comments);
		gsf_xml_out_add_bool (output, "ReplaceKeepStrings", sr->replace_keep_strings);
		gsf_xml_out_add_bool (output, "ByRow", sr->by_row);
		gsf_xml_out_add_enum (output, "Scope", GNM_SEARCH_REPLACE_SCOPE_TYPE, sr->scope);
		
		gsf_xml_out_add_int (output, "Error", sr->error_behaviour);
		/* gsf_xml_out_add_enum (output, "Error", GnmSearchReplaceError, sr->error_behaviour); */
		
		/* TODO: this int casting must be bad... */
		gsf_xml_out_add_int (output, "QueryFunc", (int)sr->query_func);
		/* gsf_xml_out_add_enum (output, "QueryFunc", GnmSearchReplaceQueryFunc, sr->query_func); */

		cmdio_write_sheet (output, NULL, sr->sheet);
		gsf_xml_out_simple_element (output, "RangeText", sr->range_text);
		
		/* TODO: void *user_data; */
		gsf_xml_out_end_element (output); /* </SearchReplace> */
				
		gsf_xml_out_end_element (output); /* </Command> */
	}

}

void
cmdio_write_colrow_std_size	(WorkbookControl *wbc, Sheet *sheet,
				 gboolean is_cols, double new_default)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_colrow_std_size\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdColRowStdSize");
		gsf_xml_out_add_bool (output, "IsCols", is_cols);
		gsf_xml_out_add_float (output, "NewDefault", new_default, -1);
		
		cmdio_write_sheet (output, NULL, sheet);

		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_zoom	(WorkbookControl *wbc, GSList *sheets, double factor)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_zoom\n");

	if (wb->change_tracker.enabled == TRUE) {
		GSList *ptr;
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdZoom");
		gsf_xml_out_add_float (output, "Factor", factor, -1);
		
		gsf_xml_out_start_element (output, "Sheets");
		for (ptr = sheets; ptr != NULL ; ptr = ptr->next) {
			cmdio_write_sheet (output, NULL, ptr->data);
		}
		gsf_xml_out_end_element (output); /* </Sheets> */
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_objects_delete	(WorkbookControl *wbc, GSList *objects,
				 char const *name)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_objects_delete\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdObjectsDelete");
		
		gsf_xml_out_start_element (output, "Objects");
		for ( ; objects != NULL; objects = objects->next) {
			cmdio_write_sheet_object (output, objects->data);
		}
		gsf_xml_out_end_element (output); /* </Objects> */
		
		gsf_xml_out_simple_element (output, "Name", name);
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_objects_move	(WorkbookControl *wbc,
				 GSList *objects, GSList *anchors,
				 gboolean objects_created, char const *name)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_objects_move\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdObjectsMove");
		gsf_xml_out_add_bool (output, "ObjectsCreated", objects_created);
		
		gsf_xml_out_start_element (output, "ObjAnch");
		for ( ; objects != NULL && anchors != NULL; objects = objects->next, anchors = anchors->next) {
			/* TODO: combine these into 1 function? */
			cmdio_write_sheet_object (output, objects->data);
			cmdio_write_anchor (output, anchors->data);
		}
		gsf_xml_out_end_element (output); /* </ObjAnch> */
	
		gsf_xml_out_simple_element (output, "Name", name);
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_object_format	(WorkbookControl *wbc, SheetObject *so,
				 gpointer orig_style)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_object_format\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdObjectFormat");
		
		cmdio_write_sheet_object (output, so);
		
		/* TODO: GObject style	=>	gpointer orig_style */
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_reorganize_sheets	(WorkbookControl *wbc,
				 WorkbookSheetState *old_state,
				 Sheet *undo_sheet)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_reorganize_sheets\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdReorganizeSheets");
		
		gsf_xml_out_start_element (output, "OldState");
		/* TODO: see workbook.c:1178
		 *
		 * gsf_xml_out_add_int (output, "NumberSheets", old_state->n_sheets);
		 * GSList old_state->properties
		 *
		 * WorkbookSheetStateSheet old_state->sheets
		 *  -	Sheet *sheet;
		 *  -	GSList *properties;
		 */
		gsf_xml_out_end_element (output); /* </OldState> */
		
		cmdio_write_sheet (output, "UndoSheet", undo_sheet);
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}

void
cmdio_write_rename_sheet	(WorkbookControl *wbc, Sheet *sheet,
				 char const *new_name)
{
	Workbook *wb = wb_control_get_workbook(wbc);
	g_print("ENTER write_rename_sheet\n");

	if (wb->change_tracker.enabled == TRUE) {
		GsfXMLOut *output = wb->change_tracker.output;
		
		gsf_xml_out_start_element (output, "Command");
		gsf_xml_out_add_cstr_unchecked (output, "Type", "CmdRenameSheet");
		
		cmdio_write_sheet (output, NULL, sheet);
		gsf_xml_out_simple_element (output, "NewName", new_name);
		
		gsf_xml_out_end_element (output); /* </Command> */
	}
}


